package xyz.veronie.bgg.result;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.eclipse.e4.core.di.annotations.Creatable;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import xyz.veronie.bgg.types.EventConstants;
import xyz.veronie.bgg.types.FamilyType;
import xyz.veronie.bgg.types.FilterFlagState;
import xyz.veronie.bgg.types.SourceFilter;
import xyz.veronie.bgg.types.Subtype;
import xyz.veronie.bgg.types.UserFlag;

@Creatable
public class BggApi {

	@Inject private ResultConfigManager configManager;
	@Inject private IEventBroker eventBroker;
	
	public static String BASE_URL = "https://www.boardgamegeek.com/xmlapi2/";
	public static String USER = "user";
	public static String COLLECTION = "collection";
	public static String SEARCH = "search";
	public static String FAMILY = "family";
	public static String GEEKLIST = "geeklist";
	
	private static int RETRIES = 5;
	
	public ArrayList<Thing> getThings(SourceFilter source, Object id) {
		switch(source) {
		default:
		case BGG_USER:
			return getThingsForUser((String)id);
		case FAMILY:
			return getThingsForFamily((Integer)id);
		case GEEKLIST:
			return getThingsForGeeklist((Integer)id);
		}
	}
	
	/// retrieve a list of things from BGG with the given ids
	public ArrayList<Thing> getThings(Set<Integer> ids) {
		StringBuilder urlStr = new StringBuilder();
		urlStr.append(BASE_URL + "thing?id=");
		
		String sep = "";
		for (Integer integer : ids) {
			urlStr.append(sep + integer.toString());
			if(sep.isEmpty()) {
				sep = ",";
			}
		}
		
		return getThings(urlStr.toString());
	}
	
	public ArrayList<Thing> getThingDetails(Set<Integer> ids) {
		StringBuilder urlStr = new StringBuilder();
		urlStr.append(BASE_URL + "thing?id=");
		
		String sep = "";
		for (Integer integer : ids) {
			urlStr.append(sep + integer.toString());
			if(sep.isEmpty()) {
				sep = ",";
			}
		}
		
		urlStr.append("&stats=1");
		
		return getThings(urlStr.toString());
	}
	

	public ArrayList<Thing> getThingsForUser(String user) throws IllegalArgumentException {
		ResultConfig resultConfig = configManager.getResultConfig();
		
		StringBuilder urlStr = new StringBuilder();
		urlStr.append(BASE_URL + COLLECTION 
				+ "?username=" + resultConfig.user
				+ "&stats=1"
				+ "&subtype=" + resultConfig.subtype.toApiString() + "&");
		if(resultConfig.subtype == Subtype.BOARDGAME) {
			urlStr.append("excludesubtype=" + Subtype.BGEXPANSION.toApiString() + "&");
		}
		
		Iterator<Entry<UserFlag, FilterFlagState>> ufIterator 
			= resultConfig.userFlags.entrySet().iterator(); 
		  
		try {
			while(ufIterator.hasNext()) {
				Entry<UserFlag, FilterFlagState> elem = ufIterator.next(); 
				String apiString = elem.getKey().toApiString();
				FilterFlagState state = elem.getValue();
				if(state == FilterFlagState.IS) {
					urlStr.append(URLEncoder.encode(apiString, "UTF-8"));
					urlStr.append("=1&");
				} else if (state == FilterFlagState.ISNOT) {
					urlStr.append(URLEncoder.encode(apiString, "UTF-8"));
					urlStr.append("=0&");
				}
			}
		}
		catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			return null;
		}

		return getThings(urlStr.toString());
	}
	
	public ArrayList<Thing> getThingsForGeeklist(int geeklistId) throws IllegalArgumentException {
		ResultConfig resultConfig = configManager.getResultConfig();
		
		StringBuilder urlStr = new StringBuilder();
		urlStr.append(BASE_URL + GEEKLIST + "/" + String.valueOf(resultConfig.geeklistId));

		return getThings(urlStr.toString());
	}
	
	public ArrayList<Thing> getThingsForFamily(int familyId) throws IllegalArgumentException {
		ResultConfig resultConfig = configManager.getResultConfig();
		
		StringBuilder urlStr = new StringBuilder();
		urlStr.append(BASE_URL + FAMILY + "?id=" + String.valueOf(resultConfig.familyId));

		if(resultConfig.familyType != FamilyType.ALL) {
			urlStr.append("?type=" + resultConfig.familyType);			
		}
		
		return getThings(urlStr.toString());
	}
	

	private ArrayList<Thing> getThings(String urlString) throws IllegalArgumentException {
		System.out.println("URL: " + urlString);
		
		try {
			int retry = 0;
			do {
				// send a GET request and check for ok...
				URL url = new URL(urlString);
				HttpURLConnection con = (HttpURLConnection) url.openConnection();
				con.setRequestMethod("GET");			
				int status = con.getResponseCode();
				if(status == HttpURLConnection.HTTP_ACCEPTED) {
					Thread.sleep(2000);
			        eventBroker.send(EventConstants.TOPIC_STATUS, "Waiting for server to prepare result...");
				}
				else if(status != HttpURLConnection.HTTP_OK) {
					eventBroker.send(EventConstants.TOPIC_STATUS, "Http Response: " + status);
					con.disconnect();
					return null;
				} else {
					// HTTP_OK, go on...		
					BufferedReader in = new BufferedReader(
							new InputStreamReader(con.getInputStream(), "UTF-8"));
					String inputLine = new String();
					StringBuffer content = new StringBuffer();
					while ((inputLine = in.readLine()) != null) {
					    content.append(inputLine);
					}
					in.close();
					con.disconnect();
					
					// do something with the content
					System.out.println(content.toString());
					ArrayList<Thing> output = parseThings(content.toString());

					output.sort((thing1, thing2) -> thing1.getMetaData().getName().compareTo(thing2.getMetaData().getName()));
					
					return output;
				}
				
				++retry;
			} while (retry < RETRIES);

		}
		catch (MalformedURLException e) {			
			eventBroker.send(EventConstants.TOPIC_STATUS, "[ERROR] Malformed URL: " + urlString);
			e.printStackTrace();
		} catch (IOException e) {
			eventBroker.send(EventConstants.TOPIC_STATUS, "[ERROR] Could not open connection for URL: " + urlString);
			e.printStackTrace();
		} catch (InterruptedException e) {
			// this will not happen, but we have to catch it anyways...
			e.printStackTrace();
		}
		
		return null;
	}

	
	
	private ArrayList<Thing> parseThings(String content) throws IllegalArgumentException {
		ArrayList<Thing> things = new ArrayList<Thing>();
		
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		try {
			
			DocumentBuilder builder = factory.newDocumentBuilder();
			InputSource is = new InputSource(new StringReader(content));
			is.setEncoding("UTF-8");
			Document doc = builder.parse(is);
			Element root = doc.getDocumentElement();
			
			checkForErrors(doc);
			
			// BGG XMLAPI2
			if(root.getTagName() == "items") {
				NodeList nList = doc.getElementsByTagName("item");
				for(int i = 0; i < nList.getLength(); i++) {
					Node nNode = nList.item(i);
					if (nNode.getNodeType() == Node.ELEMENT_NODE) {
						Element eElement = (Element) nNode;
						
						// thing when fetching by user
						if( eElement.hasAttribute("objecttype") 
								&& eElement.getAttribute("objecttype").equals("thing"))
						{
							extractThingByUser(things, eElement);						
						} 
						// when fetching things by id, type is boardgame
						else if(eElement.hasAttribute("type") &&
							eElement.getAttribute("type").equals("boardgame")) 
						{
							extractThingsById(things, nNode, eElement);				
						}			
						
						// family has "type"
						else if(eElement.hasAttribute("type") 
								&& eElement.getAttribute("type").equals("boardgamefamily")) 
						{
							extractThingByFamily(things, doc, eElement);
							
							// because of our input, there shouldn't be more than one family in the result
							// but do not iterate over items just to make sure
							break;
						}
					}
				}
			}
			// for geeklist result
			else if(root.getTagName() == "geeklist") {
				extractThingByGeeklist(things, doc);
			}

			return things;
			
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		
		
		return null;
	}

	private void extractThingByGeeklist(ArrayList<Thing> things, Document doc) {
		StringBuilder gInfo = new StringBuilder();
		
		NodeList nListTitle = doc.getElementsByTagName("title");
		if(nListTitle.getLength() > 0) {
			gInfo.append(nListTitle.item(0).getTextContent() + "\r\n");
		}

		NodeList nListDesc = doc.getElementsByTagName("description");
		if(nListDesc.getLength() > 0) {
			gInfo.append(nListDesc.item(0).getTextContent());
		}
		eventBroker.send(EventConstants.TOPIC_GEEKLIST_INFO, gInfo.toString());
		
		NodeList nList = doc.getElementsByTagName("item");
		for(int i = 0; i < nList.getLength(); i++) {
			Node nNode = nList.item(i);
			if (nNode.getNodeType() == Node.ELEMENT_NODE) {
				Element eElement = (Element) nNode;
				Integer id = Integer.parseInt(eElement.getAttribute("objectid"));
				if(id != 0) {
					ThingMetaData tmd = new ThingMetaData(
							id,
							eElement.getAttribute("objectname"), 
							"", "");
					Thing thing = new Thing(id, tmd);
					things.add(thing);
				}
			}
		}
	}

	private void extractThingByFamily(ArrayList<Thing> things, Document doc, Element eElement) {
		// get name and description
		StringBuilder fInfo = new StringBuilder();
		
		NodeList nListName = eElement.getElementsByTagName("name");
		if(nListName.getLength() > 0) {
			fInfo.append(nListName.item(0).getTextContent() + "\r\n");
		}

		NodeList nListDesc = doc.getElementsByTagName("description");
		if(nListDesc.getLength() > 0) {
			fInfo.append(nListDesc.item(0).getTextContent());
		}
		eventBroker.send(EventConstants.TOPIC_FAMILY_INFO, fInfo.toString());

		NodeList nLinks = eElement.getElementsByTagName("link");
		for(int l = 0; l < nLinks.getLength(); l++) {
			Node link = nLinks.item(l);
			if (link.getNodeType() == Node.ELEMENT_NODE) {
				Element eLink = (Element) link;
				Integer id = Integer.parseInt(eLink.getAttribute("id"));
				if(id != 0) {
					ThingMetaData tmd = new ThingMetaData(
							id,
							eLink.getAttribute("value"),
							"", "");
					Thing thing = new Thing(id, tmd);
					things.add(thing);
				}						
			}
		}
	}

	private void extractThingsById(ArrayList<Thing> things, Node nNode, Element eElement) {
		Integer id = Integer.parseInt(eElement.getAttribute("id"));
		if(id != 0) {
			String name = "";
			NodeList nameList = eElement.getElementsByTagName("name");
			for(int n = 0; n < nameList.getLength(); n++) {
				Node nameNode = nameList.item(n);
				if (nNode.getNodeType() == Node.ELEMENT_NODE) {
					Element nameElement = (Element) nameNode;
					if(nameElement.hasAttribute("type") 
							&& nameElement.getAttribute("type").equals("primary")) 
					{
						name = nameElement.getAttribute("value");
						break;
					}
				}
			}
			
			ThingMetaData tmd = new ThingMetaData(
					id,
					name,
					getValue(eElement, "image"),
					getValue(eElement, "thumbnail"));
			
			System.out.println("DEBUG: Details for " + id);
			
			ThingDetails details = new ThingDetails();
			details.yearpublished = getIntegerFromString(getValueAttribute(eElement, "yearpublished"));
			details.minplayers = getIntegerFromString(getValueAttribute(eElement, "minplayers"));
			details.maxplayers = getIntegerFromString(getValueAttribute(eElement, "maxplayers"));
			details.playingtime = getIntegerFromString(getValueAttribute(eElement, "playingtime"));
			details.minplaytime = getIntegerFromString(getValueAttribute(eElement, "minplaytime"));
			details.maxplaytime = getIntegerFromString(getValueAttribute(eElement, "maxplaytime"));
			details.age = getIntegerFromString(getValueAttribute(eElement, "minage"));
			
			NodeList statsNodes = eElement.getElementsByTagName("statistics");
			if(statsNodes.getLength() > 0) {
				Node statsNode = statsNodes.item(0);
				if (statsNode.getNodeType() == Node.ELEMENT_NODE) {
					Element statsElement = (Element) statsNode;
					NodeList ratingNodes = statsElement.getElementsByTagName("ratings");
					if(ratingNodes.getLength() > 0) {
						Node ratingNode = ratingNodes.item(0);
						if (ratingNode.getNodeType() == Node.ELEMENT_NODE) {
							Element ratingElement = (Element)ratingNode;
							details.usersrated = getIntegerFromString(getValueAttribute(ratingElement, "usersrated"));
							details.average = getFloatFromString(getValueAttribute(ratingElement, "average"));
							details.averageweight = getFloatFromString(getValueAttribute(ratingElement, "averageweight"));
							details.stddev = getFloatFromString(getValueAttribute(ratingElement, "stddev"));							
						}
					}
					
					
				}
			}
			
			
			
			// number of players recommendations
			// Only saving up to 20 players
			NodeList pollList = eElement.getElementsByTagName("poll");
			for(int n = 0; n < pollList.getLength(); n++) {
				Node pollNode = pollList.item(n);
				if (pollNode.getNodeType() == Node.ELEMENT_NODE) {
					Element pollElement = (Element) pollNode;
					if(pollElement.hasAttribute("name")) {
						String pollType = pollElement.getAttribute("name");
						if(pollType.equals("suggested_numplayers")) {
							for(int p = 0; p < 20; ++p) {
								if(p+1 < details.minplayers || p+1 > details.maxplayers) {
									details.players.add(Recommendation.N);
								} else {
									details.players.add(getMajorityRecommendation(pollElement, p+1));
								}
							}
						}
						
					}
				} // other polls
			}

			NodeList linkList = eElement.getElementsByTagName("link");
			for(int n = 0; n < linkList.getLength(); n++) {
				Node linkNode = linkList.item(n);
				if (linkNode.getNodeType() == Node.ELEMENT_NODE) {
					Element linkElement = (Element) linkNode;
					if(linkElement.hasAttribute("type")) {
						String type = linkElement.getAttribute("type");
						switch(type) {
						case "boardgamecategory":
							details.category.add(extractLinkFromElement(linkElement));
							break;
						case "boardgamemechanic":
							details.mechanic.add(extractLinkFromElement(linkElement));
							break;
						case "boardgamefamily":
							details.family.add(extractLinkFromElement(linkElement));
							break;
						case "boardgamedesigner":
							details.designer.add(extractLinkFromElement(linkElement));
							break;
						case "boardgameartist":
							details.artist.add(extractLinkFromElement(linkElement));
							break;
						case "boardgamepublisher":
							details.publisher.add(extractLinkFromElement(linkElement));
							break;
						case "boardgameexpansion":
							details.expansion.add(extractLinkFromElement(linkElement));
							break;
						case "boardgamedomain":
							details.domain.add(extractLinkFromElement(linkElement));
							break;							
						default:
							System.out.println("WARN: ignoring link element: " + type);
						}
					}
				}
			}
			
			Thing thing =  new Thing(id, tmd);
			thing.setDetails(details);
			things.add(thing);
		}
	}

	
	private IdString extractLinkFromElement(Element linkElement) {
		if(linkElement.hasAttribute("id") && linkElement.hasAttribute("value")) {
			return new IdString(Integer.valueOf(linkElement.getAttribute("id")), linkElement.getAttribute("value"));
		}
		System.out.println("WARN: Link is missing attributes.");
		return null;
	}

	private Recommendation getMajorityRecommendation(Element pollElement, int i) {
		Recommendation recommendation = null;

		NodeList nodeList = pollElement.getElementsByTagName("results");
		for(int n = 0; n < nodeList.getLength(); ++n) {
			Element elem = (Element)nodeList.item(n);
			if(elem.hasAttribute("numplayers")) {
				String attr = elem.getAttribute("numplayers");
				boolean numPlayersDoesMatch = false;
				// numplayers can be X+ where X is the max number of players... 
				if(attr.endsWith("+")) {
					attr = attr.substring(0, attr.length()-1);
					if(Integer.valueOf(attr) <= i) {
						numPlayersDoesMatch = true;
					}
				} else {
					numPlayersDoesMatch = Integer.valueOf(attr) == i;
				}

				if(numPlayersDoesMatch) {
					int lastvotes = 0;
					NodeList results = elem.getElementsByTagName("result");
					for(int r = 0; r < results.getLength(); ++r) {
						Element result = (Element)results.item(r);
						if(result.hasAttribute("value") && result.hasAttribute("numvotes")) {
							Integer numvotes = Integer.valueOf(result.getAttribute("numvotes"));
							if(numvotes > lastvotes) {
								recommendation = recommendationValueToEnum(result.getAttribute("value"));
							}
						}
						
					}
				}
			}
		}
		
		return recommendation;
	}

	private Recommendation recommendationValueToEnum(String attribute) {
		switch(attribute) {
		case "Best":
			return Recommendation.B;
		case "Recommended":
			return Recommendation.R;
		case "Not Recommended":
			return Recommendation.P;
		default:
			return Recommendation.N;
		}
	}

	private void extractThingByUser(ArrayList<Thing> things, Element eElement) {
		Integer id = Integer.parseInt(eElement.getAttribute("objectid"));
		if(id != 0) {
			ThingMetaData tmd = new ThingMetaData(
					id,
					getValue(eElement, "name"),
					getValue(eElement, "image"),
					getValue(eElement, "thumbnail")
					);
			ThingUserData tud = new ThingUserData();
			tud.setNumPlays(Integer.valueOf(getValue(eElement, "numplays")));
			tud.setComment(getValue(eElement, "comment"));

			// rating
			NodeList statsNodes = eElement.getElementsByTagName("stats");
			if(statsNodes.getLength() > 0) {
				Node statsNode = statsNodes.item(0);
				if (statsNode.getNodeType() == Node.ELEMENT_NODE) {
					Element statsElement = (Element) statsNode;
					NodeList ratingNodes = statsElement.getElementsByTagName("rating");
					if(ratingNodes.getLength() > 0) {
						Element ratingNode = (Element)ratingNodes.item(0);
						String ratingValue = ratingNode.getAttribute("value");
						if(!ratingValue.equals("N/A")) {
							try {
								Float rating = Float.valueOf(ratingValue);
								tud.setRating(rating);
							}
							catch(NumberFormatException e) {
								System.out.println("WARN: rating value is not a float."); 
							}
						}
					}
					
					
				}
			}
			
			NodeList statusNodes = eElement.getElementsByTagName("status");
			if(statusNodes.getLength() > 0) {
				Element statusNode = (Element)statusNodes.item(0);
				tud.setOwn(attributeToBoolean(statusNode, "own"));
				tud.setPrevowned(attributeToBoolean(statusNode, "prevowned"));
				tud.setFortrade(attributeToBoolean(statusNode, "fortrade"));
				tud.setWant(attributeToBoolean(statusNode, "want"));
				tud.setWanttoplay(attributeToBoolean(statusNode, "wanttoplay"));
				tud.setWanttobuy(attributeToBoolean(statusNode, "wanttobuy"));
				tud.setWishlist(attributeToBoolean(statusNode, "wishlist"));
				if(statusNode.hasAttribute("wishlistpriority")) {
					tud.setWishlistpriority(Integer.valueOf(statusNode.getAttribute("wishlistpriority")));
				}
				tud.setPreordered(attributeToBoolean(statusNode, "preordered"));
				tud.setLastmodified(statusNode.getAttribute("lastmodified"));
			}
			
			Thing thing =  new Thing(id, tmd);
			thing.setUserData(tud);
			things.add(thing);
		}
	}


	private Boolean attributeToBoolean(Element statusNode, String attribute) {
		if(statusNode.hasAttribute(attribute)) {
			return statusNode.getAttribute(attribute).equals("0") ? false : true;
		} 
		return null;
	}
	
	private void checkForErrors(Document doc) throws IllegalArgumentException {
		NodeList nList = doc.getElementsByTagName("error");
		if(nList.getLength() > 0) {
			Element el = (Element)nList.item(0);
			String val = getValue(el, "message");
			if(val.equals("")) {
				val = el.getAttribute("message");
			}
			throw new IllegalArgumentException(val);
		}
	}

	private Integer getIntegerFromString(String s) {
		if(s != null && !s.isEmpty()) {
			return Integer.valueOf(s);
		} 
		return null;
	}

	private Float getFloatFromString(String s) {
		if(s != null && !s.isEmpty()) {
			return Float.valueOf(s);
		} 
		return null;
	}

	private String getValue(Element eElement, String key) {
		Node node = eElement.getElementsByTagName(key).item(0);
		if(node == null) {
			return "";
		} else {
			return node.getTextContent();
		}        
	}
	
	private String getValueAttribute(Element eElement, String key) {
		Node node = (Element)eElement.getElementsByTagName(key).item(0);
		if (node != null && node.getNodeType() == Node.ELEMENT_NODE) {
			Element elem = (Element) node;
			if(elem.hasAttribute("value")) {
				return elem.getAttribute("value");
			}
		}
		return null;
	}
	
}
